import { getContextItem } from "../../lib/context.ts";
import { TrackSystem } from "../../track/system.ts";
import { Debuggable } from "../../lib/debuggable.ts";
import { PhysicsSolver } from "../../physics/solver.ts";
import { MoveTask, TrackConstraint, TrainTask } from "./physics.ts";
import { TrainCar } from "./TrainCar.ts";
import { Bogie } from "./Bogie.ts";

export class Train extends Debuggable {
  private _bogies: Bogie[];
  get bogies() {
    return this._bogies;
  }

  private solver: PhysicsSolver;

  constructor(private track: TrackSystem, private cars: TrainCar[] = []) {
    super("train");
    this._bogies = this.recalculateBogies();
    this.solver = new PhysicsSolver();
    this._bogies.forEach((b) =>
      this.solver.addTask(new TrackConstraint(b, this.track))
    );
    let prev: Bogie | null = null;
    for (const car of this.cars) {
      for (const b of car.bogies) {
        if (prev) {
          this.solver.addTask(
            new TrainTask(prev, b, prev.trailingOffset + b.leadingOffset),
          );
        }
        prev = b;
      }
    }
    this.solver.addTask(new MoveTask(this));
    this.solver.maxTickStep = 1 / 30;
  }

  addCar(car: TrainCar, toFront = false) {
    this.cars.push(car);
    for (const b of toFront ? car.bogies.reverse() : car.bogies) {
      this.solver.addTask(new TrackConstraint(b, this.track));
    }
    this._bogies = this.recalculateBogies();
  }
  removeCar(car: TrainCar) {
    this.cars = this.cars.filter((c) => c !== car);
  }

  recalculateBogies() {
    return this.cars.flatMap((c) => c.bogies);
  }

  update(dt: number) {
    this.solver.solve(dt);
  }

  draw() {
    for (const car of this.cars) {
      car.draw();
    }
  }

  debugDraw(): void {
    const d = getContextItem("doodler");
    // this.track._segments.forEach((s) =>
    //   s.evenPoints.forEach((p) => d.dot(p[0], { color: "lime", weight: 2 }))
    // );
  }
}
